from collections.abc import Sequence
import inspect
from mongoengine import Document
from graphene_mongodb.operators import list_fields, reference_fields
from graphene_mongodb.fields.respective import respective_fields, respective_special_fields


class Options:
    """ This class makes all necessary verifications of types that was given by user """

    def __init__(self, class_name, attrs):
        self.class_name = class_name
        self.model, self.mutate, self.mongo_fields, self.validator = self.verified_attrs(attrs)

    def verified_attrs(self, attrs):
        """ Function to verify if the attributes is of the right type """
        model = attrs.get('model')
        mutate = attrs.get('mutate')
        validator = attrs.get('validator')
        validator = validator.__func__ if isinstance(validator, staticmethod) else validator

        if model is None:
            raise AttributeError('Failed to generate schema {},'
                                 ' model attribute was not given.'.format(self.class_name))

        if not model or not issubclass(model, Document):
            raise TypeError('Failed to generate schema {}, model must be '
                            'a subclass of mongoengine.Document.'.format(self.class_name))

        if mutate and not isinstance(mutate, staticmethod):
            raise TypeError('Failed to generate schema {}, mutate method must '
                            'be a method with the decorator staticmethod.'.format(self.class_name))

        if mutate and len(inspect.signature(mutate.__func__).parameters) != 2:
            raise TypeError('Failed to generate schema {}, mutate method must accept two params. '
                            'The first is the arguments passed to mutate in query, for instance: username:"NewObjName".'
                            ' Second is the context of the application, if it is flask, will be flask global request.'
                            .format(self.class_name))

        # verify if all fields of document is supported
        for f_name, m_field in model._fields.items():
            if type(m_field) not in respective_fields and type(m_field) not in respective_special_fields:
                raise NotImplementedError("It was not possible to generate schema for {} because the "
                                          "field {} is of the type {}, and that field is not supported yet."
                                          .format(model.__name__, f_name, type(m_field)))

            if isinstance(m_field, list_fields):
                if type(m_field.field) not in respective_fields and type(m_field.field) not in respective_special_fields:
                    raise NotImplementedError("It was not possible to generate schema for {} because the "
                                              "field {} is a List of the type {}, and that field is not supported yet."
                                              .format(model.__name__, f_name, type(m_field.field)))

            if isinstance(m_field, reference_fields):
                if isinstance(m_field.document_type_obj, str) and m_field.document_type_obj == 'self':
                    raise NotImplementedError("It was not possible to generate schema for {} because the "
                                              "field {} is a ReferenceField to self and this is not supported yet."
                                              .format(self.class_name, f_name))

        if validator and not callable(validator):
            raise AttributeError("'validator' attribute must be callable.")
        elif validator and len(inspect.signature(validator).parameters) != 4:
            raise AttributeError("The 'validator' attribute must be a callable that accepts four arguments: "
                                 "model, fields, query, special_params")

        return model, mutate.__func__ if mutate else None, model._fields, validator



import graphene
from graphene.utils.str_converters import to_snake_case

from graphene_mongodb.query import mongo_to_graphene


def gen_mutation(model, graphene_schema, operators_mutation, fields_mutation, mutate_func, validator):
    """ We need to create a class that seems as follows (http://docs.graphene-python.org/en/latest/types/mutations/):

    class CreatePerson(graphene.Mutation):
        class Input:
            name = graphene.String()
    
        ok = graphene.Boolean()
        person = graphene.Field(lambda: Person)
    
        @staticmethod
        def mutate(root, args, context, info):
            person = Person(name=args.get('name'))
            ok = True
            return CreatePerson(person=person, ok=ok) 
    """

    def user_mutate(root, info, **kwargs):
        if validator:
            validator(model, kwargs, {}, {})

        obj = mutate_func(kwargs, info.context)
        if not isinstance(obj, model):
            raise TypeError('Failed to resolve mutation of the schema {}'
                            ' because mutate function must return a instance of {}, and the return type was {}.'
                            .format(graphene_schema.__name__, model.__name__, type(obj)))

        graphene_obj = mongo_to_graphene(obj, graphene_schema, fields_mutation)
        return Create(**{to_snake_case(model.__name__): graphene_obj})

    def generic_mutate(root, info, **kwargs):
        if validator:
            validator(model, kwargs, {}, {})

        obj = model(**kwargs)
        obj.save()
        graphene_obj = mongo_to_graphene(obj, graphene_schema, fields_mutation)
        return Create(**{to_snake_case(model.__name__): graphene_obj})

    Create = type('Create' + model.__name__, (graphene.Mutation,), {
        'Arguments': type('Arguments', (), operators_mutation),
        to_snake_case(model.__name__): graphene.Field(lambda: graphene_schema),
        'mutate': staticmethod(generic_mutate) if not mutate_func else staticmethod(user_mutate)
    })

    return Create



import graphene


def reference_field(f_name, mongo_field):
    """ Generate a schema for RefereceField, or get a schema already done saved in _generated_schemas """
    from graphene_mongodb import MongoSchema

    document = mongo_field.document_type_obj  # document that this ReferenceField references

    schema = MongoSchema.get_or_generate_schema(document).schema
    return graphene.Field(schema)


def list_field(f_name, mongo_field):
    from graphene_mongodb import MongoSchema
    from graphene_mongodb.fields.respective import respective_fields

    list_items_type = type(mongo_field.field)

    if list_items_type in respective_fields:
        return graphene.List(type(respective_fields[list_items_type]()))
    else:
        try:
            document = mongo_field.field.document_type
        except AttributeError:
            raise AttributeError('Error in {} field, have sure that this is defined with a mongoengine field'
                                 .format(f_name))

        schema = MongoSchema.get_or_generate_schema(document).schema
        return graphene.List(schema)



from six import with_metaclass
from graphene_mongodb.model import ModelSchema
from graphene_mongodb.options import Options


class MongoSchemaMeta(type):
    def __new__(cls, *args):
        class_name, parents, attrs = args

        if not parents:
            return type.__new__(cls, class_name, parents, attrs)

        options = Options(class_name, attrs)
        generated_schema = cls._gen_schema(options.model, options)
        options.single = generated_schema.single
        options.list = generated_schema.list

        return type(class_name, (), dict(generated_schema.to_attrs(), _meta=options))

    _generated_schemas = {}

    @classmethod
    def _gen_schema(cls, model, options=None):
        """ This function is responsible for generate the graphene schema and memoize it """
        options = Options(model.__name__ + 'Schema', {'model': model}) if not options else options

        MongoSchemaMeta._generated_schemas.update({
            model: ModelSchema(model, options.mongo_fields, options.mutate, options.validator)
        })
        return MongoSchemaMeta._generated_schemas[model]

    @classmethod
    def get_or_generate_schema(cls, model):
        if model not in cls._generated_schemas:
            return cls._gen_schema(model)
        else:
            return cls._generated_schemas.get(model)


class MongoSchema(with_metaclass(MongoSchemaMeta)):
    def __new__(cls, model):
        return type(model.__name__ + 'Schema', (MongoSchema,), {'model': model})




import graphene

from graphene_mongodb.fields import convert_fields
from graphene_mongodb.query import resolver_query
from graphene_mongodb.mutation import gen_mutation


class ModelSchema:
    def __init__(self, model, attrs_mongo_doc, mutate, validator):
        self.model = model
        self.mutate_func = mutate  # mutate function that user specified
        self.validator = validator

        self.fields, self.fields_mutation, self.operators_mutation,\
            self.operators_single, self.operators_list = convert_fields(attrs_mongo_doc)

        self.schema = type(self.model.__name__ + 'Graphene', (graphene.ObjectType,), self.fields.copy())

    def to_attrs(self):
        return {
            'single': self.single(),
            'list': self.list(),
            'fields': self.fields,
            'model': self.model,
            'mutate': self.mutate().Field()
        }

    def mutate(self):
        return gen_mutation(self.model, self.schema, self.operators_mutation, self.fields_mutation, self.mutate_func,
                            self.validator)

    def single(self):
        return ModelSchema.resolver(self.schema, self.model, operators_single=self.operators_single,
                                    validator=self.validator)

    def list(self):
        return ModelSchema.resolver(self.schema, self.model, operators_list=self.operators_list,
                                    validator=self.validator, is_list=True)

    @staticmethod
    def resolver(g_schema, mongo_doc, operators_single=None, operators_list=None, is_list=False, validator=None):
        def auto_resolver(root, info, **kwargs):
            return resolver_query(g_schema, mongo_doc, kwargs, info, is_list=is_list, validator=validator)

        if is_list:
            return graphene.List(g_schema, **operators_list, resolver=auto_resolver)
        else:
            return graphene.Field(g_schema, **operators_single, resolver=auto_resolver)



import graphene
from graphene.types.datetime import DateTime
from mongoengine import StringField, BooleanField, IntField, FloatField, DateTimeField, ObjectIdField, URLField, \
    DictField, EmailField, LongField, DecimalField, BinaryField, PointField, ReferenceField, ListField, SortedListField

from graphene_mongodb.fields.custom_fields import GenericField, CustomBinaryField, CustomDecimalField
from graphene_mongodb.fields.special_fields import reference_field, list_field


# http://docs.mongoengine.org/guide/defining-documents.html?highlight=emailfield#fields
respective_fields = {
    StringField: graphene.String,
    BooleanField: graphene.Boolean,
    IntField: graphene.Int,
    FloatField: graphene.Float,
    DateTimeField: DateTime,
    ObjectIdField: graphene.ID,
    URLField: graphene.String,
    DictField: GenericField,
    EmailField: GenericField,
    LongField: GenericField,
    DecimalField: CustomDecimalField,
    BinaryField: CustomBinaryField,
    PointField: GenericField,
}

respective_special_fields = {
    ReferenceField: reference_field,
    ListField: list_field,
    SortedListField: list_field
}


def field_to_id(m_field, g_field):
    """ We need this because if we want to do a query using the id, we will pass a string to args with the id of the 
    document that we want, but graphene needs a ID field, instead of Field. This function convert to right thing."""

    if isinstance(m_field, ReferenceField):
        return graphene.ID()
    elif (isinstance(m_field, ListField) or isinstance(m_field, SortedListField)) and \
            isinstance(m_field.field, ReferenceField):
        """ Pass here if it is a ListField or SortedListField of ReferenceField """
        return graphene.List(graphene.ID)
    else:
        return g_field



import graphene
from mongoengine import ObjectIdField

from graphene_mongodb.query import special_query_parameters
from graphene_mongodb.operators import gen_operators_of_field, allowed_operators
from graphene_mongodb.fields.respective import respective_special_fields, respective_fields, field_to_id


def convert_fields(attrs_mongo_doc):
    """ Return a tuple od dicts,
    @param attrs_mongo_doc: dict of attributes of the mongoengine.Document, that is accessible by Document._fields     
    fields: is the pure fields, key is the name of field, and values is the repective graphene field
    fields_mutation: is the fields that is used in mutation, that will go direct in args of mutation
    operators_mutation: is the fields that user set values to be saved on document
    operators_single: is the fields with operators that will be in the query for a single result
    operators_list: same of operators single, but these are avaliable only for query of lists
    """

    fields, fields_mutation, operators_mutation, operators_single = {}, {}, {}, {}

    for f_name, mongo_field in attrs_mongo_doc.items():
        field = RelationMongoGraphene(name=f_name, mongo_field=mongo_field)

        fields[f_name] = field.graphene
        operators_single.update(field.operators)
        fields_mutation[f_name] = field.mutation
        if not isinstance(mongo_field, ObjectIdField):
            operators_mutation[f_name] = field.mutation

    operators_list = operators_single.copy()
    operators_single['skip'] = graphene.Int()

    operators_list.update({p: graphene.Int() for p in special_query_parameters})

    return fields, fields_mutation, operators_mutation, operators_single, operators_list


class RelationMongoGraphene:
    """ Abstraction of Relation between mongoengine field and graphene field """

    def __init__(self, name, mongo_field):
        self.name = name
        self.mongo = mongo_field
        self.graphene = RelationMongoGraphene.to_respective_graphene(name, mongo_field)

        self.mutation = field_to_id(self.mongo, self.graphene)
        self.operators = gen_operators_of_field(self.name, self.mongo, self.graphene, allowed_operators(self.mongo))

    @staticmethod
    def to_respective_graphene(f_name, mongo_field):
        """ Function that given field, returns the respective field of graphene """
        t_mongo_field = type(mongo_field)

        if t_mongo_field in respective_fields:
            return respective_fields[t_mongo_field]()
        else:
            return respective_special_fields[t_mongo_field](f_name, mongo_field)



import graphene
from mongoengine import IntField, FloatField, DateTimeField, LongField, DecimalField, BooleanField, ObjectIdField, \
                        DictField, BinaryField, PointField, ListField, SortedListField, ReferenceField, StringField, \
                        URLField, EmailField
from graphene_mongodb.fields.respective import field_to_id

# http://docs.mongoengine.org/guide/querying.html#query-operators
operators = {
    'ne': lambda m_field, g_field:  field_to_id(m_field, g_field),
    'lt': lambda m_field, g_field:  field_to_id(m_field, g_field),
    'lte': lambda m_field, g_field: field_to_id(m_field, g_field),
    'gt': lambda m_field, g_field:  field_to_id(m_field, g_field),
    'gte': lambda m_field, g_field: field_to_id(m_field, g_field),
    'in': lambda m_field, g_field:  graphene.List(type(field_to_id(m_field, g_field))),
    'nin': lambda m_field, g_field: graphene.List(type(field_to_id(m_field, g_field))),
    'exists': lambda _, __:         graphene.Boolean(),
    'size': lambda _, __:           graphene.Int()
}

# http://docs.mongoengine.org/guide/querying.html#string-queries
string_operators = ['exact', 'iexact', 'contains', 'icontains', 'startswith', 'istartswith', 'endswith', 'iendswith']

fields_string_operators = (StringField, URLField, EmailField)
reference_fields = (ReferenceField,)
scalar_fields = (IntField, FloatField, DateTimeField, LongField, DecimalField)
specific_fields = (BooleanField, ObjectIdField, DictField, BinaryField, PointField)
list_fields = (ListField, SortedListField)


def allowed_operators(mongo_field):
    if isinstance(mongo_field, fields_string_operators) or isinstance(mongo_field, scalar_fields):
        return list(operators.keys())
    elif isinstance(mongo_field, specific_fields):
        return ['ne', 'in', 'nin', 'exists']
    elif isinstance(mongo_field, reference_fields):
        return ['in', 'nin', 'ne']
    elif isinstance(mongo_field, list_fields):
        return ['size']


def gen_operators_of_field(f_name, mongo_field, r_graphene, operators_list):
    """ Return a dict with keys as the name of the field with operator and value is the required type, for instance: 
    @param f_name: string name of the field
    @param mongo_field: object instance of mongoengine field, e.g: mongoengine.StringField()
    @param r_graphene: object instance of graphene field, e.g: graphene.String(): 
    {
        name: graphene.String()
        name__nin: graphene.List(graphene.String) ...
    }
     """

    field_with_operators = {
        f_name: field_to_id(mongo_field, r_graphene)
    }

    for op_name in operators_list:
        field_with_operators[f_name + '__' + op_name] = operators[op_name](mongo_field, r_graphene)

    if isinstance(mongo_field, fields_string_operators):
        for op in string_operators:
            field_with_operators[f_name + '__' + op] = graphene.String()

    return field_with_operators



import iso8601
from graphql.utils.ast_to_dict import ast_to_dict


def parse_operators(args):
    """ Avoid problem that mongoengine for some reason the operators 
        gte, gt, lt, lte doesn't work with dates in isoformat 
    """

    args = {k: v for k, v in args.items() if k not in ['skip', 'limit']}
    for k, v in args.items():
        try:
            is_data = iso8601.parse_date(v)
        except:
            is_data = False

        if is_data:
            args[k] = is_data if is_data else v

    return args


def mongo_to_graphene(mongo_obj, graphene_obj, fields):
    """ Pass fields from mongoengine object to graphene """
    return graphene_obj(**{f: getattr(mongo_obj, f) for f in fields})


# author: mixxorz
def collect_fields(node, fragments):
    field = {}

    if node.get('selection_set'):
        for leaf in node['selection_set']['selections']:
            if leaf['kind'] == 'Field':
                field.update({
                    leaf['name']['value']: collect_fields(leaf, fragments)
                })
            elif leaf['kind'] == 'FragmentSpread':
                field.update(collect_fields(fragments[leaf['name']['value']], fragments))

    return field


# author: mixxorz
def get_fields(info):
    fragments = {}
    node = ast_to_dict(info.field_asts[0])

    for name, value in info.fragments.items():
        fragments[name] = ast_to_dict(value)

    return collect_fields(node, fragments)


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name='graphene_mongodb',
  packages=find_packages(exclude=['tests', 'tests.*', 'examples', 'flask_example']),
  version='1.0.1',
  license='MIT',
  description='GrapheneMongo is a library that integrates Graphene with MongoEngine',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='João Vitor Silvestre',
  author_email='joaovitorsilvestresousa@gmail.com',
  url='https://github.com/joaovitorsilvestre/graphene-mongo',
  download_url='https://github.com/joaovitorsilvestre/graphene-mongo/archive/1.0.1.tar.gz',
  keywords=['graphene', 'graphql', 'mongo', 'mongodb', 'mongoDB', 'mongoengine', 'graphene-mongo', 'graphene-mongodb'],
  classifiers=[],
  install_requires=[
    'aniso8601>=3.0.0',
    'appdirs>=1.4.3',
    'graphene>=2.1',
    'graphql-core>=2.0',
    'graphql-relay>=0.4.5',
    'iso8601>=0.1.12',
    'mongoengine>=0.15.0',
    'promise>=2.1',
    'pymongo>=3.6.1',
    'pyparsing>=2.2.0',
    'rx>=1.6.1',
    'six>=1.11.0',
    'typing>=3.6.4'
  ]
)


from graphene.utils.str_converters import to_snake_case
from graphene_mongodb.query.utils import get_fields, parse_operators, mongo_to_graphene


# Options that we can pass as operators to define the query
special_query_parameters = ['skip', 'limit']


def do_query(m_object, query, fields, special_params, is_list):
    query = m_object.objects(**query).only(*fields)

    if 'skip' in special_params:
        query = query.skip(special_params['skip'])
    if 'limit' in special_params:
        query = query.limit(special_params['limit'])

    return query


def resolver_query(g_object, m_object, args, info, is_list=False, validator=None):
    fields = [to_snake_case(f) for f in get_fields(info)]
    query = parse_operators(args)
    special_params = {k: v for k, v in args.items() if k in ['skip', 'limit']}

    if validator:
        validator(m_object, fields, query, special_params)

    result = do_query(m_object, query, fields, special_params, is_list)

    if result and is_list:
        return [mongo_to_graphene(obj, g_object, fields) for obj in result]
    elif result:
        return mongo_to_graphene(result.first(), g_object, fields)
    else:
        return [] if is_list else None



from graphene.types import Scalar
from graphql.language.ast import FloatValue, IntValue


class GenericField(Scalar):
    @staticmethod
    def serialize(_info):
        return _info


class CustomDecimalField(Scalar):
    @staticmethod
    def coerce_float(value):
        try:
            return float(value)
        except ValueError:
            return None

    serialize = coerce_float
    parse_value = coerce_float

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, (FloatValue, IntValue)):
            return float(ast.value)


class CustomBinaryField(Scalar):
    @staticmethod
    def serialize(_binary):
        return _binary.decode('utf-8')
